#define _FILE_OFFSET_BITS 64

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
// #include <queue>
#include <string.h>
// #include <vector>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>


const char * invocation_name;

const int blocksize = 512;  // block size (one sector)
const int max_blocks_per_rd = 64;  // maximum number of blocks (sectors) to read at a time
typedef char sector_t[512];
sector_t bufr[64];
static sector_t zerosector;


int main( const int argc, const char * argv[] ) {
  long long ipos = 0, opos = 0;
  long num_blocks;
  long num_blocks_left;
  long blocks2rd, rdblocks, rdsize;
  long num_wrote = 0, num_skipped = 0;
  int percentdone = 0;
  int i, final_partial_block_size;

  invocation_name = argv[0];

  if ((3 != argc) || (!argv[1]) || (!argv[2]))
    {
      fprintf( stderr, "sparsecopy -- sparsely copy a file, only copying non-0 sectors.\n"
                       "Output is overwritten w/o truncation (for merging ddrescue image files).\n"
                       "\n"
                       "Usage:\n"
                       "  %s inputfile outputfile\n"
                       "\n"
                       "by Dave Burton, www.burtonsys.com (uncopyrighted)\n", invocation_name );
      return 1;
    }
  const char * iname = argv[1];
  const char * oname = argv[2];
  // fprintf( stderr, "dbg: 1\n" ); fflush(stderr);
  if( !strcmp( iname, oname ) )
    {
      fprintf( stderr, "ERROR: input name and output name cannot be identical.\n" );
      return 1;
    }

  // fprintf( stderr, "dbg: 2\n" ); fflush(stderr);
  const int ides = open( iname, O_RDONLY );
  if( ides < 0 )
    {
      fprintf( stderr, "ERROR: could not open input file '%s', errno=%d\n", iname, errno );
      return 1;
    }

  // fprintf( stderr, "dbg: 3\n" ); fflush(stderr);
  const long long isize = lseek( ides, 0, SEEK_END );
  if( isize < 0 )
    {
      printf( "ERROR: input file '%s' is not seekable.", iname );
      return 1;
    }
  lseek( ides, 0, SEEK_SET );
  ipos = 0;

  // fprintf( stderr, "dbg: 4\n" ); fflush(stderr);
  const int odes = open( oname, O_WRONLY | O_CREAT, 0644 );
  if( odes < 0 )
    {
      fprintf( stderr, "ERROR: cannot open output file '%s', errno=%d\n", oname, errno );
      return 1;
    }
  // fprintf( stderr, "dbg: 5\n" ); fflush(stderr);
  if( lseek( odes, 0, SEEK_SET ) )
    {
      printf( "ERROR: output file '%s' is not seekable.", oname );
      return 1;
    }
  opos = 0;

  // fprintf( stderr, "dbg: 6\n" ); fflush(stderr);

  num_blocks_left = num_blocks = isize / blocksize;
  final_partial_block_size = isize - (num_blocks * blocksize);
  if (final_partial_block_size)
    {
      fprintf( stderr, "Note: '%s' is not an even multiple of %d bytes in length.\n",
               iname, blocksize ); fflush(stderr);
    }
  
  fprintf( stderr, "Sparsely copying %lld bytes (%ld blocks) from '%s' to '%s'\n",
           isize, num_blocks, iname, oname );  fflush(stderr);
  
  percentdone = 0;
  // fprintf( stderr, "dbg: 7\n" ); fflush(stderr);
  while (num_blocks_left > 0)
    {
      long tmp = (((num_blocks - num_blocks_left) * 4) + 2) / (num_blocks / 25);
      if (tmp > percentdone)
        {
          percentdone = tmp;
          fprintf( stderr, "\r %2d%% \r", percentdone );
          fflush( stderr );
        }

      // read max_blocks_per_rd blocks at a time
      if (num_blocks_left > max_blocks_per_rd) {
          blocks2rd = max_blocks_per_rd;
      } else {
          blocks2rd = num_blocks_left;
      }
      rdsize = read( ides, bufr, blocksize * blocks2rd );
      rdblocks = rdsize / blocksize;
      if ((blocksize * blocks2rd) != rdsize)
        {
          fprintf( stderr, "ERROR: attempted read of %ld sectors from '%s', but only got %ld.\n", blocks2rd, iname, rdblocks );
          if (rdblocks > blocks2rd)
            {
              printf( "ERR: read too much -- should be impossible!\n" );
              return 1;
            }
          if ((rdsize <= blocksize) || ((rdblocks * blocksize) != rdsize))
            {
              printf( "ERROR: stopping because can't read from input file '%s'.\n", iname );
              return 1;
            }
        }

      // write just 1 block at a time
      for (i=0; i<rdblocks; i++)
        {
          if (memcmp( bufr[i], zerosector, blocksize))
            {
              opos = ipos + (i * blocksize);
              if (opos != lseek( odes, opos, SEEK_SET ) )
                {
                  fprintf( stderr, "ERR: lseek failed on output file.  Aborting.\n" );
                  return 1;
                }
              if (blocksize != write( odes, bufr[i], blocksize ))  // write( bufr[i], blocksize, 1, odes ))
                {
                  fprintf( stderr, "ERR: write to output file failed.  Aborting.\n" );
                  return 1;
                }
              num_wrote++;
            }
          else
              num_skipped++;
        }

      ipos += (rdblocks * blocksize);
      num_blocks_left -= rdblocks;
    }//while
    
  if (final_partial_block_size)
    {
      fprintf( stderr, "Note: '%s' is not an even multiple of %d bytes in length.\n", iname, blocksize ); fflush(stderr);
      rdsize = read( ides, bufr[0], final_partial_block_size );
      if (rdsize != final_partial_block_size)
        {
          fprintf( stderr, "ERROR: reading final %d bytes from '%s', only read %ld bytes.\n", final_partial_block_size, iname, rdsize );
          return 1;
        }
      if (memcmp( bufr[0], zerosector, rdsize))
        {
          opos = ipos;
          if (opos != lseek( odes, opos, SEEK_SET ) )
            {
              fprintf( stderr, "ERR: final lseek failed on output file.  Aborting.\n" );
              return 1;
            }
          if (rdsize != write( odes, bufr[0], rdsize ))  // write( bufr[i], blocksize, 1, odes ))
            {
              fprintf( stderr, "ERR: final write to output file failed.  Aborting.\n" );
              return 1;
            }
          fprintf( stderr, "Final %d bytes were not all zeros, so they were written to '%s'.\n",
                   final_partial_block_size, oname );
        }
      else
        {
          fprintf( stderr, "Final %d bytes were not written to '%s' because they're all zeros.\n",
                   final_partial_block_size, oname );
        }
       ipos += rdsize;
    }

  fprintf( stderr, "Done.  %ld sectors copied, %ld zero-sectors skipped.\n", num_wrote, num_skipped );
  if ((num_wrote+num_skipped) != num_blocks)
    {
      fprintf( stderr, "ERR: %ld + %ld != %ld.  Program is broken!\n", num_wrote, num_skipped, num_blocks );
    }
  return 0;
} //main
